#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include "basic.h"

static const int vmax = 4 ;
#define NALL (vmax*vmax)
static int verbose = 0 ;


/* Barrier */


#include "sense_barrier.c"

typedef uint64_t tb_t ;
#define PTB PRIu64

static tb_t rdtsc(void) {
  uint32_t a,d ;
  asm __volatile__ ("rdtsc" : "=a" (a), "=d" (d) :: "memory");
  return ((tb_t)a) | (((tb_t)d)<<32);
}

typedef struct {
  int size ;
  barrier_t *b ;
  volatile uint64_t *x,*y;
  volatile uint64_t *r1,*r2;
  tb_t *t1,*t2;
} common_t ;

typedef struct {
  int id ;
  common_t *common ;
} ctx_t ;


void *T1(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {    
    wait_my_barrier(q->b,&sense) ;
    for (int i = 1 ; i < vmax ; i++) {
      q->x[k] = i ;
      q->t1[k] = rdtsc() ;
    }
    int r1 = q->y[k] ;
    q->r1[k] = r1 ;
  }
  return NULL ;
}

void *T2(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;
  int sense = 0 ;
  int size = q->size ;
  for (int k = size-1 ; k >= 0 ; k--) {
    wait_my_barrier(q->b,&sense) ;
    for (int i = 1 ; i < vmax ; i++) { 
      q->y[k] = i ;
      q->t2[k] = rdtsc() ;
    }
    int r2 = q->x[k] ;
    q->r2[k] = r2 ;
  }
  return NULL ;
}


typedef struct {
  int size ;
  volatile int stop ;
  pthread_mutex_t *mutex ;
} run_common_t ;

typedef struct {
  int id ;
  run_common_t *common;
} run_t ;


void *run(void *_p) {
  run_t *p = (run_t *)_p ;
  run_common_t *pc = p->common;
  int size = pc->size ;

  barrier_t *b = alloc_my_barrier(2) ;
  uint64_t count[vmax][vmax] ;
  for (int i = 0 ; i < vmax ; i++)
    for (int j = 0 ; j < vmax ; j++)
      count[i][j] = 0 ;

  uint64_t *x = calloc(size,sizeof(x[0])) ;
  uint64_t *y = calloc(size,sizeof(y[0])) ;
  uint64_t *r1 = calloc(size,sizeof(r1[0])) ;
  uint64_t *r2 = calloc(size,sizeof(r2[0])) ;
  tb_t *t1 = calloc(size,sizeof(t1[0])) ;
  tb_t *t2 = calloc(size,sizeof(t2[0])) ;
  common_t c ;
  c.size = size ; c.b = b ; c.x = x ; c.y = y ;
  c.r1 = r1 ; c.r2 = r2 ; c.t1 = t1 ; c.t2 = t2 ;

  for ( ; !pc->stop ; ) {
    pthread_t th1,th2 ;
    ctx_t a1,a2 ;
    
    for (int k = 0 ; k < size ; k++) {
      x[k] = y[k] = 0 ;
      r1[k] = r2[k] = 0 ;
      t1[k] = t2[k] = 0 ;
    }
    a1.id = 1 ; a1.common = &c ;
    create_thread(&th1,T1,&a1) ;
    a2.id = 2 ; a2.common = &c ;
    create_thread(&th2,T2,&a2) ;

    (void)join_thread(&th1) ;
    (void)join_thread(&th2) ;
    for (int k = 0 ; k < size ; k++) {
      count[r1[k]][r2[k]]++ ;
    }

    int nz = 0 ;
    for (int k1 = 0 ; k1 < vmax ; k1++)
      for (int k2 = 0 ; k2 < vmax ; k2++) {
        uint64_t c = count[k1][k2] ;
        if (c > 0) nz++ ;
      }
    if (verbose || nz == NALL) {
      lock_mutex(pc->mutex) ;
      printf("**** %i ****\n",p->id) ;
      for (int k1 = 0 ; k1 < vmax ; k1++)
        for (int k2 = 0 ; k2 < vmax ; k2++) {
        uint64_t c = count[k1][k2] ;
        printf("%-8" PRIu64 "> r1=%i r2=%i\n",c,k1,k2) ;
      }
      if (!verbose && nz == NALL) pc->stop = 1 ;
      unlock_mutex(pc->mutex) ;      
    }
  }
  free(x) ; free(y) ; free(r1) ; free(r2) ;
  free(t1) ; free(t2) ;
  free_my_barrier(b) ;
  return NULL;
}

void zyva(int size,int n) {
  run_common_t c ;
  c.size = size ;
  c.stop = 0 ;
  c.mutex = alloc_mutex() ;
  pthread_t th[n] ;
  run_t arg[n] ;
  for (int k = 0 ; k < n ; k++) {
    arg[k].id = k ;
    arg[k].common = &c ;
    create_thread(&th[k],run,&arg[k]) ;
  }
  for (int k = 0 ; k < n ; k++) (void)join_thread(&th[k]) ;
}

void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 100000 ;
  int n = 4 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc >= 2) {    
    size = atoi(argv[1]) ;
  }
  if (argc >= 1) {    
    n = atoi(argv[0]) ;
  }
  zyva(size,n) ;
  return 0 ;
}
